; Benchmarking tools 
; version1.2.0
; Fri 02 Apr 2004 11:03:28 AM MSD
;
; procedures for creating random DB and reading it from file
;
; Exported: 
;
; facts-count
;
; create-db
; read-idb-edb
; read-db
; read-object
; 
;
; benchmark
; bench-query
; 
;=============================================================================
; Useful

; facts-count :: db -> integer
; returns summarize count of facts in the db
(define (facts-count db)
  (fold (lambda (set res) (+ (length set) res)) 0 db))

;=============================================================================
; Procedures

; create-db :: db-info -> filename -> unspecified
; db-info is information about db. It's a list of lists.
; Each list consist of names of relation, arity of relation, 
; max count of facts and max-int-value.
; facts consist of constants from 0 to max-int-value.
; create-db write the random db in the file filename.
(define (create-db db-info filename)
  (if (file-exists? filename)
    (delete-file filename))
  (let* ((db (instantiate::sqlite (path filename))))
  (for-each 
    (lambda (cur-db-info)
      (let* ((name (car cur-db-info))
             (arity (value->number (cadr cur-db-info)))
             (max-power (value->number (caddr cur-db-info)))
	     (max-int-value (value->number (cadddr cur-db-info)))
	     (sql (concat "create table " name " ( "))
	     (ins-q (concat "insert into " name " values (~k)")))
      (do ((i 0 (+ i 1)))
          ((= i arity))
	  (set! sql (concat sql "f" (number->string i) " varchar(20)"))
	  (if (= i (- arity 1)) (set! sql (concat sql ")"))
				(set! sql (concat sql ", "))))
      (sqlite-exec db sql)
      (for-each 
        (lambda (tuple)
	  (sqlite-exec db ins-q tuple))
        (delete-duplicates
          (let loop-tuples ((i max-power))
           (if (= i 0)
            '()
            (cons (let iter ((i 0))
                    (if (= i arity)
                        '()
                        (cons (number->string (random max-int-value)) (iter (+ i 1)))))
                  (loop-tuples (- i 1))))))))) db-info)))

; read-db :: filename -> full-db
; filename is a file contains SQLite database.
; All tables from this database are written to full-db. (with names)
(define (read-db filename)
  (if (not (file-exists? filename))
      (begin (cout "Database file doesn't exist" nl) (exit -1)))
  (define askq "select * from sqlite_master")
  (let* ((db (instantiate::sqlite (path filename))))
  (let ((return
    (map 
      (lambda (schema-tuple)
        (let* ((tablename (second schema-tuple))
               (selq (concat "select * from " tablename)))
          (list (string->symbol tablename) 
	        (sqlite-map db vector selq))))
      (filter (lambda (schema-tuple) (equal? (first schema-tuple) "table")) 
          (sqlite-map db list askq) ))))
   (sqlite-close db)
   return)))


; insert-values -> tablename -> values -> conn -> sess -> unspecified
; values is a list of values
;(define (insert-values tablename values conn sess)
;  (define sql (concat "insert into " tablename " values ( "))
;  (define len (length values))
;  (cout "insert-values " values nl)
;  (do ((i 0 (+ i 1)))
;          ((= i len))
;    (set! sql (concat sql ":" (number->string (+ i 1))))
;    (if (= i (- len 1)) (set! sql (concat sql ")"))
;			(set! sql (concat sql ", "))))
;  (cout "sql = " sql nl)
;  (prepare sess sql)
;  (bind! sess values)
;  (execute sess)
;  (cout "ok" nl)
; )
			
; read-idb-edb :: filename -> idb-names -> [edb-names idb edb]
; This is a function on the read-db function.
; If we know idb-names (only after program's analyze), we can divide our
; database (db = start-idb + edb) to idb and edb and can know edb-names
(define (read-idb-edb filename idb-names)
  (define db (read-db filename))
;  (cout "db = " db nl)
  (let* ((edb-full
           (filter (lambda (x) (if (not (member (car x) idb-names)) x #f)) db))
	 (idb (map 
        	(lambda (idb-name) 
  	          (let* ((val (find (lambda (x) (equal? (car x) idb-name)) db)))
	            (if val (second val) '()))) idb-names)))
  (receive (edb-names edb) (unzip2 edb-full)
    (values edb-names idb edb))))

; read-object :: filename -> object (rules, ra-eqs, etc)
; read-object read object from filename
(define (read-object filename)  
  (define port (open-input-file filename))
  (let ((res (read port)))  
    (close-input-port port)
    res))

; this program output the line:
; Yb-time GZ-time CI-time QSQ-time facts-edb inferred-facts 
;   and for each query  relevant-facts Yb-time GZ-time CI-time QSQ-time

; benchmark :: methods -> program -> idb-names -> edb-names -> edb -> [idb list]
; list is a pair [name time] measurements for each method of methods
; one query (the first) in program is need by QSQR   
;(define (benchmark methods program idb-names edb-names edb)
; (define ticks (clocks-per-second))
; (define rules (cadr program))
; (define ra-eqs (dtlog->ra rules))
; (define query (cdar (third program)))
; (define gl-idb #f)
; (define (measure method)
;   (define name (car method))
;   (receive (total utime stime cutime cstime) (times)
;   (let* ((proc (second method))
;          (idb 
;	    (if (equal? name 'QSQR)
;                (list (qsqr rules idb-names edb-names edb query))
;                (proc ra-eqs idb-names edb-names edb)))) 
;     (receive (ntotal nutime nstime ncutime ncstime) (times)
;     (if (not gl-idb) (set! gl-idb idb))
;     (cout (/ (- nutime utime) ticks) "  ")
;     (list name (/ (- nutime utime) ticks))))))
; (let* ((count-edb (facts-count edb))
;        (mes (remove (lambda (method) (equal? (car method) 'QSQR))
;               (map measure methods))))
;  (cout count-edb "  ")
;  (cout (facts-count gl-idb) "   ")
;  (values gl-idb mes))
;)

;bench-query :: program -> idb-names -> edb-names -> edb -> idb -> mes-times -> unspecified
; such as benchmark proc, surves to benchmark all queries in program
; exclude first query. 
; mes-times is a result of benchmark call
;(define (bench-query program idb-names edb-names edb idb mes-times)
; (define ticks (clocks-per-second))
; (define rules (cadr program))
; (define ra-eqs (dtlog->ra rules))
; (define queries (cdr (map cdr (third program))))
; (define (measure name query)
;   (receive (total utime stime cutime cstime) (times)
;   (let ((result
;	(if (equal? name 'QSQR)
;            (list (qsqr rules idb-names edb-names edb query))
;	    (begin (open-query query edb-names idb-names edb idb #f) idb))))
;     (receive (ntotal nutime nstime ncutime ncstime) (times)
;     (list (facts-count result) (/ (- nutime utime) ticks))))))
; (for-each (lambda (query) 
;        (let* ((qsqr-mes (measure 'QSQR query))
;	       (ra-mes (measure #f query)))
;	  (cout (car qsqr-mes) "  ")
;	  (for-each (lambda (mes) 
;    	      (cout (+ (second mes) (second ra-mes)) "  ")) mes-times)
;	  (cout (second qsqr-mes) "   ")))
;         queries))

