;; Calculating methods of Datalog
;; Mon 07 Apr 2003 09:21:05 PM MSD
;; version1.0.0 7.04.2003
;  
; Relational equation (rule) is represented as s-expression.


; Exported:
;
; qsqr 
; op-ra
;
;==============================================================================
; Calculation methods of Datalog
; choice :: body of rule -> next sub aim 
; returns first (next) sub aim of the rule body
(define (choice rule-body)
  (car rule-body))

; op-bindings :: sub-aim -> rule-proc -> sel-cond and join-cond
; returns two list - conditionals to select and join operation for determine
; what ra-operation we need to process sub-aim and rule-proc
(define (op-bindings sub-aim rule-proc)
  (let rpt ((terms (cdr sub-aim)) (proc-terms '()) (index 0))
    (if (null? terms)
        (values '() '())
        (let ((term (car terms)))
	  (receive (sel-cond join-cond)
	    (rpt (cdr terms) (append proc-terms (list term)) (+ index 1))
          (if (var? term)
              (let ((j-index (rbody-index rule-proc term))
	            (s-index (list-index (lambda(x) (equal? x term)) proc-terms)))
		(values
		  (if s-index 
		      (append sel-cond (list (list index 
		      ; this list of (index, lambda) is a truk for or-and-sel-expr. index is not used
		        (lambda (elem) (equal? (vector-ref elem index) 
		                               (vector-ref elem s-index))))))
		      sel-cond)
                  (if j-index 
	              (append join-cond (list (list j-index index)))
		      join-cond)))
              (values (append sel-cond (list (list index term))) join-cond)))))))

; const-bind :: sub-aim -> rule-proc -> %temp-res -> c-query
; return common query, where all known bindings of variables are substituted
; to the constants of %temp-res
(define (const-bind sub-aim rule-proc %temp-res)
  (map (lambda (x) 
         (if (var? x)
	     (let ((index (rbody-index rule-proc x)))
	       (if index (map (lambda (y) 
	                        (vector-ref y index)) %temp-res)
	    	   x)) x)) sub-aim))

 ; add-query :: q -> query -> [new-q boolean]
; adds query to set q consists of common-queries, 
; returns new-q and #t if query is new indeed and #f otherwise
(define (add-query q query)
 (if (any (lambda (x) (equal? x query)) q)
     (values q #f)
 (let ((new-q
  (let rpt ((q q))
    (if (null? q)
        (list query)
        (let* ((q-elem (car q))
               (eq-pos (eq-var-pos? q-elem query))
               (query (if eq-pos (car (rename q-elem (list query)))))
               (mgu-s (if eq-pos (mgu (only-vars query) (only-vars q-elem)) #f)))
;             (cout q-elem nl eq-pos nl query nl  nl mgu-s nl)
          (if mgu-s 
	      (append 
	        (list (map (lambda (x y) 
	               (if (const? x)
	                   (delete-duplicates 
		             (append (const->set x) (const->set y))) x))
		     q-elem query)) (cdr q))
	      (cons q-elem (rpt (cdr q)))))))))
  (if (equal? new-q q) 
      (values q #f)
      (values new-q #t)))))
;(cout (add-query '((sgc X X "a" ) ) '(sgc X Y "b")) nl)
;=> '((sgc X X (a b)))
;(cout (add-query '((sgc X X ("a" "b")) ) '(sgc X Y "b")) nl)
;=> #f

; op-ra :: sub-aim -> processed rule -> ra operation
; return new ra operation = lambda of %temp idb and %predicate
(define (op-ra sub-aim rule-proc)
; (if (eq? (car sub-aim) eq-name)
;   (lambda (%temp %pred)
;     (if (var? (second sub-aim)) (third sub-aim) (second sub-aim)))
 (receive (sel-cond join-cond) (op-bindings sub-aim rule-proc)
  (lambda (%temp %pred)
    ;(cout "op-ra " sel-cond nl)
    (let ((f-sel (or-and-sel-expr sel-cond)))
      (if (null? rule-proc)
          (select %pred f-sel)
          (join %temp (select %pred f-sel)
                      (and-join-expr join-cond)))))))
;(define %temp ((op-ra '(par X A) '()) '() (car edb)))
;(cout ((op-ra '(par A Y) '((par X A))) (car edb) (car edb)) nl)

; query->c-queries :: query -> c-queries
; transforms query into list of sub-queries
(define (query->c-queries query)
  (map (lambda (x) 
         (map (lambda (y) (if (const? y) 
	                      (list y)
			      y)) x)) query)) 	  

; fold-idb :: pred-name -> result -> idb-names -> idb -> [new-idb boolean]
; accumulate result of pred-name in idb, returns new-idb
; boolean is #f if new-idb equals idb
(define (fold-idb pred-name result idb-names idb)
;  (cout "fold-idb" nl "idb = " idb nl "result = " result nl)
  (receive (before after)
    (split-at idb (list-index (lambda (x) (eq? x pred-name)) idb-names))
;    (cout before nl after nl)
    (receive (new-idb-pred is-new)
      (union-info (car after) result)
;      (cout (car after) nl result nl new-idb-pred nl)
      (values (append before (cons new-idb-pred (cdr after))) is-new))))
;(trace union-info)
;(receive (new-idb bool)
;  (fold-idb 'sgc '(#(1 2)) '(sgc2 sgc) '(() (#(1 2) #(0 1))) )
; => '(() (#(1 2) #(0 1))) and #f
;  (fold-idb 'sgc '(#(1 3)) '(sgc2 sgc) '(() (#(1 2) #(0 1))) )  
; => '(() (#(1 3) #(1 2) #(0 1))) and #t
; (fold-idb '%anc '(#2("b" "a") #2("b" "g") #2("a" "d") #2("a" "e") #2("d" "f") #2("c" "h"))
;     '(%anc) '(()))
; => ((#(b a) #(b g) #(a d) #(a e) #(d f) #(c h))) and #t
;  (fold-idb '%anc '(#2("b" "d") #2("b" "e") #2("a" "f"))
;      '(%anc) '((#2("b" "a") #2("b" "g") #2("a" "d") #2("a" "e") #2("d" "f") #2("c" "h"))))
; => ((#(a f) #(b e) #(b d) #(b a) #(b g) #(a d) #(a e) #(d f) #(c h)))      
;  (cout new-idb nl bool nl))

; open-query :: query -> edb-names -> idb-names -> edb -> idb -> qsq-f -> result or [result new-idb]
; open-query calculate query on idb consists of edb-names
;                           and ebd consists of idb-names
; qsq-f is a function that is called when current predicate is idb-predicate.
; qsq-f :: idb -> new-query -> [new-idb idb-changed?]
; qsq-f is recursive call is needed to qsq method.
; if recursive call isn't needed qsq-f must be #f and open-query returns only result
; otherwise it returns both result and new-idb
(define (open-query query edb-names idb-names edb idb qsq-f)
 (define (result %temp idb)
   (if qsq-f (values %temp idb) %temp))
 (let rpt ((rule-unproc query) (rule-proc '()) (%temp '()) (idb idb))
;    (cout "rpt:" nl rule-unproc "  " rule-proc nl %temp nl)	    
    (if (null? rule-unproc) 
        (result %temp idb)
	(let ((sub-aim (car rule-unproc)))
	  (if (eval-expr? sub-aim) ; we suggest all eval exprs are at the end of rule
	      (rpt '() '() (process-eval-exprs (flat-rule-terms rule-proc) %temp #f rule-unproc (append edb idb) (append edb-names idb-names))
    	    	    idb)
              (let* ((op (op-ra sub-aim rule-proc))
	        (pred (car sub-aim))
	        (edb-pred? (list-index (lambda (x) (eq? x pred)) edb-names))
	        (db-ind (if edb-pred? edb-pred?
		                   (list-index (lambda (x) (eq? x pred)) idb-names))))
; 		  (cout "rpt processed sub-aim:" sub-aim nl)
                     (let* ((idb (if qsq-f
                		  (if (or edb-pred? (eq? pred eq-name))
               			   idb
                 		   (receive (idb bool) 
				     (qsq-f idb (const-bind sub-aim rule-proc %temp))
;				     (cout "qsqr: idb = " idb nl)
				     idb)) idb))
		          (%pred (cond (edb-pred? (list-ref edb db-ind))
			               ((eq? pred eq-name) (eq-result (const-bind sub-aim rule-proc %temp)))
			               (else (list-ref idb db-ind))))
		          (%res (op %temp %pred)))
;        		(cout "%pred="  %pred nl "%temp="%temp nl)
;			(cout "res:" %res nl)
			(if (eq-set? %res '())
			  (result '() idb)
                          (rpt (cdr rule-unproc) (append rule-proc (list sub-aim)) 
			   %res idb)))))))))
;(trace open-query)
 
(define (qsqr rules idb-names edb-names edb query)
; (cout "query = " query nl)
 (define (qsqr q idb idb-changed? c-query)
  (receive (q new-q?) (add-query q c-query)
; (cout nl "qsqr " q nl "IDB: " idb nl)
; (cout "new-q? = " new-q? " idb-changed? = " idb-changed? nl nl)
 (if (or new-q? idb-changed?) ; dynamic condition of breaking
  (begin (set! idb-changed? #f)
  (values (fold (lambda (rule idb) ; for - rules
;    (cout nl "idb = " idb nl)
      (let* ((c-rule (rename c-query rule))
           (mgu-subst (mgu c-query (car c-rule))))
      (if mgu-subst
      (receive (%rule-ans idb)
        (open-query (cdr (apply-rule-subst c-rule mgu-subst)) 
                    edb-names idb-names edb idb 
   	            (lambda (idb query) (qsqr q idb idb-changed? query)))
        (let ((%rule-result (projection %rule-ans (corr-list c-rule))))
   (receive (new-idb new-idb?)
     (fold-idb (caar rule) %rule-result idb-names idb)
     (set! idb-changed? (or idb-changed? new-idb?))
;     (cout "new-idb? = " new-idb? nl)
     new-idb))) idb))) ;end of fold's lambda
    idb rules) idb-changed?)) (values idb #f)) ; end of dynamic condition of breaking
    ))
 (let ((q (query->c-queries query)))
   (let repeat ((idb (make-list (length idb-names) '())))
     (receive (idb idb-changed?)
       (let my-map ((idb idb) (idb-changed? #f) (q q))
         (if (null? q)
     	     (values idb idb-changed?)
	     (receive (idb new-idb-changed?)
               (qsqr q idb #t (car q))
;	       (cout "new-idb-changed? = " new-idb-changed? nl)
;	        (cout "idb = " idb nl)
               (my-map idb (or idb-changed? new-idb-changed?) (cdr q)))))
;      (cout "DEBUG: idb = " idb nl)
       (if (not idb-changed?)
           (open-query query edb-names idb-names edb idb #f)
           (repeat idb))))))
;(trace fold-idb)
;(trace const-bind)
;(trace add-query)

;(define edb '( (#("john"   "henry")
;                #("george" "jim")
;                #("henry"  "mary")
;                #("jim"    "mary"))
;               (#("john"   "john")
;	        #("henry"  "henry")
;		#("mary"   "mary"))))
;(define rules '( ((sgc X Y) (eq X Y))
;                 ((sgc X Y) (par X X1) (sgc X1 Y1) (par Y Y1)) ))
;(cout (qsqr rules '(sgc) '(par eqdb) edb '((sgc "john" X)) ) nl)
;(cout (qsqr rules '(sgc) '(par) edb '((sgc "john" X)) ) nl)
;(cout (qsqr rules '(sgc) '(par eqdb) edb '((sgc Y X)) ) nl)

;(define rules '( ((sgc X Y) (par X A) (par A Y)) ))
;(define edb '( (#("1" "0")
;                #("2" "0")
;                #("0" "3")
;               #("1" "5"))))

;(cout (qsqr rules '(sgc) '(par) edb '((sgc ("1") X)) ))
; => (#(1 3))



;==============================================================================
; Examples

; common-query:
; '(sgc ("john") X)
; query: 
; '( '() (sgc "john" X))
; program
; '( (sgc X Y) (eq X Y))
; '( (sgc X Y) (par X X1) (sgc X1 Y1) (par Y Y1))

